/* BCD-Uhr */

#include <inttypes.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>
#include <avr/signal.h>
#include <avr/io.h>
#include <avr/delay.h>

#define TMC8_CK			(_BV(CS00))
#define TMC8_CK8		(_BV(CS01))
#define TMC8_CK64       (_BV(CS00)+_BV(CS01))
#define TMC8_CK256		(_BV(CS02))
#define TMC8_CK1024		(_BV(CS00)+_BV(CS02))

#define MAXMODES		 2				// maximale Anzahl Displaymodi
#define KEYPADDELAY		64				// Wartezeit
#define MODE			 4				// linke Taste
#define SET				 6				// rechte Taste
// fr Uhr mit nur einem Taster MODE und SET tauschen

volatile unsigned char hz=0;      		// Stunden Zehner
volatile unsigned char he=1;         	// Stunden Einer
volatile unsigned char mz=0;      		// Minuten Zehner
volatile unsigned char me=1;         	// Minuten Einer
volatile unsigned char sz=0;      		// Sekunden Zehner
volatile unsigned char se=1;         	// Sekunden Einer
volatile unsigned char stelle=0;    	// Displaystelle
volatile unsigned char takt=0;      	// 1/255 Sekundentakt
volatile unsigned char alarmstate=0;	// Alarmstatus an/aus
volatile unsigned char alarmactive=0;	// Alarm aktiv/nicht_aktiv
volatile unsigned char ahz=0; 			// Alarmzeit Stunden Zehner
volatile unsigned char ahe=6; 			// Alarmzeit Stunden Einer
volatile unsigned char amz=0; 			// Alarmzeit Minuten Zehner
volatile unsigned char ame=5; 			// Alarmzeit Minuten Einer
volatile unsigned char displaymode=1;	// Displaymodus
volatile unsigned char keypress=0xff;	// Tastencode
volatile unsigned char setmode=1;		// zu stellende Stelle

SIGNAL(SIG_OVERFLOW0)
{
if (displaymode == 0) PORTD = 0xFF;

if (displaymode == 2)
	{
	PORTD = 0;
	PORTB &= ~(_BV(PB2)+_BV(PB1)+_BV(PB0));
	_delay_loop_1(42);
	}
PORTB |= _BV(PB0)+_BV(PB1)+_BV(PB2);	// alle Stellen aus
if (setmode < 4)
{
if ((displaymode==1) || (displaymode==2))
	{
	PORTD = 0xFF;
	switch (stelle)
		{
		case 0:
		// Stunden ausgeben
		PORTD = ~((hz<<4)+he);
		if (!((setmode==1)&&(takt&64))) PORTB &= ~(_BV(PB2));  // Abfrage fr Blinken, 16 fr flackern
		break;
		case 1:
		// Minuten ausgeben
		PORTD = ~((mz<<4)+me);
		if (!((setmode==2)&&(takt&64))) PORTB &= ~(_BV(PB1));
		break;
		case 2:       
		// Sekunden ausgeben
		PORTD = ~((sz<<4)+se);
		if (!((setmode==3)&&(takt&64))) PORTB &= ~(_BV(PB0));
		break;
		}
	if ((++stelle) > 2) stelle=0;			// Stellenzhler erhhen bzw. zurcksetzen
	}
}
else										// Alarmzeit anzeigen
	{
//	PORTB |= _BV(PB0)+_BV(PB1)+_BV(PB2);	// alle Stellen aus
	switch (stelle)
		{
		case 0:
		// Stunden ausgeben
		PORTD = ~((ahz<<4)+ahe);
		if (!((setmode==4)&&(takt&16))) PORTB &= ~_BV(PB2);
		break;
		case 1:
		// Minuten ausgeben
		PORTD = ~((amz<<4)+ame);
		if (!((setmode==5)&&(takt&16))) PORTB &= ~_BV(PB1);
		break;
		case 2:       
		// Sekundenstelle ausgeben (in diesem Fall Alarm on/off)
		PORTD = ~alarmactive;
		if (!((setmode==6)&&(takt&16))) PORTB &= ~_BV(PB0);
		break;
		}
	if ((stelle++) > 2) stelle=0;			// Stellenzhler erhhen bzw. zurcksetzen
	}

keypress = PINB;

/* Zeit zhlen */
if ((++takt)==255)
	{
	se++;
	if (se>9)
		{
		se=0;
		sz++;
	 	if (sz>5)
			{
			sz=0;
			me++;
			if (me>9)
				{
				me=0;
				mz++;
				if (mz>5)
					{
					mz=0;
					he++;
					if (he>9)
						{
						he=0;
						hz++;
						}
					if (((hz*10)+he)>23)
						{
						hz=0;
						he=0;
						}
					}
				}
			}  
		}   
	}
if ((alarmstate) && (takt&128) && (takt&16)) TCCR1A |= _BV(COM1A0);
else TCCR1A &= ~_BV(COM1A0);
}

void warte(void)
{	_delay_loop_2(60000);
	_delay_loop_2(60000);
	_delay_loop_2(60000);
}

int main(void)
{
/* Registerkonfiguration */
DDRB=0x0F;								// PortB: 0..3 out; 4..7 in
DDRD=0xFF;								// PortD: all out
PORTB=0xFF;								// all off / PullUp activated
PORTD=0xFF;								// all off
TIMSK =(_BV(TOIE0));					// enable TCNT0 overflow interrupt
TCCR0 = TMC8_CK64;      				// use CLK/64 prescale
TCCR1B = (_BV(CTC1)+_BV(CS10));
OCR1A=0x0C00;
sei();

/* endlosschleife */
while(1)
	{
		if bit_is_clear(keypress,MODE)
			{
			setmode++;
			if (setmode > 6) setmode = 0;
			warte();
			}
		
		if bit_is_clear(keypress,SET)
			{
			switch (setmode) 
				{
				case 0:
					if (alarmstate) alarmstate=0;
					else
						{
						displaymode++;
						if (displaymode > MAXMODES) displaymode = 0;
						}
					break;
				case 1:
					he++;
					if (he>9)
						{
						he=0;
						hz++;
						}
					if (((hz*10)+he)>23)
						{
						hz=0;
						he=0;
						}
					break;
				case 2:
					me++;
					if (me>9)
						{
						me=0;
						mz++;
						if (mz>5)
							{
							mz=0;
							me=0;
							}
						}
					break;
				case 3:
					se=0;
					sz=0;
					setmode=0;
					break;
				case 4:
					ahe++;
					if (ahe>9)
						{
						ahe=0;
						ahz++;
						}
					if (((ahz*10)+ahe)>23)
						{
						ahz=0;
						ahe=0;
						}
					break;
				case 5:
					ame++;
					if (ame>9)
						{
						ame=0;
						amz++;
						if (amz>5)
							{
							amz=0;
							ame=0;
							}
						}
					break;
				case 6:
					alarmactive = !alarmactive;
					break;
				}
			warte();
			}
	if ((amz==mz)&&(ame==me)&&(ahz==hz)&&(ahe==he)&&(sz==0)&&(se==0)&&(alarmactive)) alarmstate=1;
	}
}
// end.
